/* $Id:$
***************************************************************************

   X server for LibGGI - Handling of graphics output

   Copyright (C) 1997 Jason McMullan	[jmcc@cs.cmu.edu]
   Copyright (C) 1997 Michael Krause	[rawstyle@ms.demo.org]
   Copyright (C) 1998 Marcus Sundberg	[marcus@ggi-project.org]

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
*/

#include <stdlib.h>
#include "X11/Xos.h"

#define PSZ 8
#include "cfb.h"

#include "scrnintstr.h"
#include "servermd.h"
#include "colormapst.h"
#include "mipointer.h"

#include "xggi.h"

#define MAX_COLORS 256 /* For the PseudoColor modes.. */

ggi_visual_t MASTER_VIS;
struct xggiScreenInfo xggiScreen =
{
	DEFAULT_PTRBUTTONS,	/* Default number of pointer buttons */
	GGI_AUTO,		/* Default width */
	GGI_AUTO,		/* Default height */
	0			/* Default depth */
};

static int
xggiOpenConsole(void)
{
	static int isopen = 0;
    
	if (isopen) return 1;

	if (ggiInit() < 0 ||
	    (MASTER_VIS = ggiOpen(NULL)) == NULL) {
		MASTER_VIS = NULL;
		return 0;
	}

#if 0	/* Can we hook in ggiFlush() somewhere? */
	ggiSetFlags(MASTER_VIS, GGIFLAG_ASYNC);
#endif
	isopen = 1;
	
	return 1;
}

void
ddxGiveUp()
{
	if (MASTER_VIS != NULL) {
		ggiClose(MASTER_VIS);
		MASTER_VIS = NULL;
		ggiExit();
	}
}

void
OsVendorInit()
{
}

void
OsVendorFatalError()
{
	ggiExit();
}

void
ddxUseMsg()
{
	ErrorF("XGGI adds the following arguments:\n");
	ErrorF("-ggimode W[xH[xD] ]    set width, height, depth (bpp)\n");
	ErrorF("-buttons nr-buttons    specify number of pointer buttons\n");
}

int
ddxProcessArgument (int argc, char *argv[], int i)
{
	if (!strcmp(argv[i], "-ggimode")) {
		if (i + 1 >= argc) {
			ErrorF("Option `-ggimode' requires an argument\n");
			return 1;
		} else if (sscanf(argv[i+1], "%dx%dx%d",
				  &xggiScreen.width,
				  &xggiScreen.height,
				  &xggiScreen.depth) < 1) {
			ErrorF("Invalid mode spec `-ggimode %s'"
			       " - using defaults\n", argv[i+1]);
		}
		return 2;
	}
	if (!strcmp(argv[i], "-buttons")) {
		if (i + 1 >= argc) {
			ErrorF("Option `-buttons' requires an argument\n");
			return 1;
		}
		xggiScreen.ptrbuttons = atoi(argv[i+1]);
		return 2;
	}

	return 0;
}

#ifdef DDXTIME /* from ServerOSDefines */
CARD32
GetTimeInMillis()
{
	struct timeval  tp;

	X_GETTIMEOFDAY(&tp);
	return TVAL2TIME(tp);
}
#endif

static ColormapPtr InstalledMaps[MAXSCREENS];

static int
xggiListInstalledColormaps(ScreenPtr pScreen, Colormap *pmaps)
{
	/* By the time we are processing requests, we can guarantee that there
	 * is always a colormap installed */
	*pmaps = InstalledMaps[pScreen->myNum]->mid;
	return (1);
}


static void
xggiInstallColormap(ColormapPtr pmap)
{
	int index = pmap->pScreen->myNum;
	ColormapPtr oldpmap = InstalledMaps[index];

	if (pmap != oldpmap) {
			int entries;
			VisualPtr pVisual;
			Pixel *     ppix;
			xrgb *      prgb;
			xColorItem *defs;
			int i;

			if (oldpmap != (ColormapPtr)None)
				WalkTree(pmap->pScreen, TellLostMap, (char *)&oldpmap->mid);
			/* Install pmap */
			InstalledMaps[index] = pmap;
			WalkTree(pmap->pScreen, TellGainedMap, (char *)&pmap->mid);

			entries = pmap->pVisual->ColormapEntries;
			pVisual = pmap->pVisual;

			ppix = (Pixel *)ALLOCATE_LOCAL(entries * sizeof(Pixel));
			prgb = (xrgb *)ALLOCATE_LOCAL(entries * sizeof(xrgb));
			defs = (xColorItem *)ALLOCATE_LOCAL(entries * sizeof(xColorItem));

			for (i = 0; i < entries; i++)  ppix[i] = i;
			/* XXX truecolor */
			QueryColors(pmap, entries, ppix, prgb);

			for (i = 0; i < entries; i++) { /* convert xrgbs to xColorItems */
				defs[i].pixel = ppix[i] & 0xff; /* change pixel to index */
				defs[i].red = prgb[i].red;
				defs[i].green = prgb[i].green;
				defs[i].blue = prgb[i].blue;
				defs[i].flags =  DoRed|DoGreen|DoBlue;
			}
			(*pmap->pScreen->StoreColors)(pmap, entries, defs);

			DEALLOCATE_LOCAL(ppix);
			DEALLOCATE_LOCAL(prgb);
			DEALLOCATE_LOCAL(defs);
		}
}

static void
xggiUninstallColormap(ColormapPtr pmap)
{
	ColormapPtr curpmap = InstalledMaps[pmap->pScreen->myNum];

	if (pmap == curpmap) {
		if (pmap->mid != pmap->pScreen->defColormap) {
			curpmap = (ColormapPtr)
				LookupIDByType(pmap->pScreen->defColormap,
					       RT_COLORMAP);
			(*pmap->pScreen->InstallColormap)(curpmap);
		}
	}
}

static void
xggiStoreColors(ColormapPtr pmap, int ndef, xColorItem  *pdefs)
{
	int i;
	ggi_color cmap[MAX_COLORS];

	if (pmap != InstalledMaps[pmap->pScreen->myNum]) {
		return;
	}

	if ((pmap->pVisual->class | DynamicClass) == DirectColor) {
		return;
	}

	ggiGetPalette(MASTER_VIS, 0, MAX_COLORS, cmap);

	for (i = 0; i < ndef; i++)
		{
			if (pdefs[i].pixel >= MAX_COLORS)
				continue;

			if (pdefs[i].flags & DoRed)
				cmap[pdefs[i].pixel].r=pdefs[i].red;
			if (pdefs[i].flags & DoGreen)
				cmap[pdefs[i].pixel].g=pdefs[i].green;
			if (pdefs[i].flags & DoBlue)
				cmap[pdefs[i].pixel].b=pdefs[i].blue;
		}

	ggiSetPalette(MASTER_VIS, 0, MAX_COLORS, cmap);
}

static Bool
xggiSaveScreen(ScreenPtr pScreen, int on)
{
#if 0
	switch (what) {
	case SCREEN_SAVER_ON:
		break;
      
	case SCREEN_SAVER_OFF:
		break;
      
	case SCREEN_SAVER_FORCER:
		break;
      
	case SCREEN_SAVER_CYCLE:
		break;
	}
#endif

	return TRUE;
}

static Bool
xggiCloseScreen(int index, ScreenPtr pScreen)
{
#if 0 /* This causes a segmentation fault on exit. Is it really necessary? */
	if ((PixmapPtr)pScreen->devPrivate) {
		return pScreen
			->DestroyPixmap((PixmapPtr)pScreen->devPrivate);
	}
#endif
	return TRUE;
}

static char *
xggiAllocateFramebufferMemory(void)
{
	const ggi_directbuffer *dbuf = NULL;
	int size, i;

	if (xggiScreen.pfbMemory) {
		return xggiScreen.pfbMemory; /* already done */
	}

	if (ggiSetMode(MASTER_VIS, &xggiScreen.mode)) {
		ErrorF("LibGGI can not set any modes at all!\n");
		return NULL;
	}
	
	size = GT_SIZE(xggiScreen.mode.graphtype);
	for (i = 0; (dbuf = ggiDBGetBuffer(MASTER_VIS, i)) != NULL; i++) {
		if ((dbuf->type & GGI_DB_SIMPLE_PLB)
		    && ((8*dbuf->buffer.plb.stride) % size) == 0) {
			xggiScreen.pfbMemory = dbuf->write;
			xggiScreen.stride = 8 * dbuf->buffer.plb.stride
				/ size;
			xggiScreen.dbuf = dbuf;
			break;
		}
	}
		    
	if (!xggiScreen.pfbMemory) {
		ErrorF("This mode and target has no suitable DirectBuffer\n");
		return NULL;
	}

	return xggiScreen.pfbMemory;
}


static Bool
xggiCursorOffScreen (ScreenPtr *ppScreen, int *x, int *y)
{
	return FALSE;
}

static void
xggiCrossScreen (ScreenPtr pScreen, Bool entering)
{
}

static miPointerScreenFuncRec xggiPointerCursorFuncs =
{
	xggiCursorOffScreen,
	xggiCrossScreen,
	miPointerWarpCursor
};

static Bool
xggiFBInitProc(int index, ScreenPtr pScreen,
	       int argc, char *argv[])
{
	const int dpix = 100, dpiy = 100;
	char *pbits;

	if (!(pbits = xggiAllocateFramebufferMemory())) {
		return FALSE;
	}

	if (!xggiScreenInit(pScreen, pbits, xggiScreen.width, xggiScreen.height,
			   dpix, dpiy, xggiScreen.stride)) {
		return FALSE;
	}

	pScreen->CloseScreen = xggiCloseScreen;
	pScreen->SaveScreen = xggiSaveScreen;

	switch(xggiScreen.depth) {
	case 15:
	case 16:
	case 24:
	case 32:
		pScreen->InstallColormap = cfbInstallColormap;
		pScreen->UninstallColormap = cfbUninstallColormap;
		pScreen->ListInstalledColormaps = cfbListInstalledColormaps;
		pScreen->StoreColors = (void (*)())NoopDDA;
		break;
	default:
		pScreen->InstallColormap = xggiInstallColormap;
		pScreen->UninstallColormap = xggiUninstallColormap;
		pScreen->ListInstalledColormaps = xggiListInstalledColormaps;
		pScreen->StoreColors = xggiStoreColors;
		break;
	}

	miDCInitialize(pScreen, &xggiPointerCursorFuncs);

	return cfbCreateDefColormap(pScreen);
} 


void
InitOutput(ScreenInfo *screenInfo, int argc, char *argv[])
{
	ggi_graphtype gt;

	if (!xggiOpenConsole()) {
		FatalError("Unable to init LibGGI visual!\n");
	}

	switch (xggiScreen.depth) {
	case 0:  gt = GT_AUTO;  break;
	case 8:  gt = GT_8BIT;  break;
	case 15: gt = GT_15BIT; break;
	case 16: gt = GT_16BIT; break;
	case 24: gt = GT_24BIT; break;
	case 32: gt = GT_32BIT; break;
	default:
		FatalError("XGGI doesn't support %d depth screens.\n",
			   xggiScreen.depth);
	}

	ggiCheckGraphMode(MASTER_VIS, xggiScreen.width, xggiScreen.height,
			  GGI_AUTO, GGI_AUTO, gt, &xggiScreen.mode);
		
	xggiScreen.width  = xggiScreen.mode.visible.x;
	xggiScreen.height = xggiScreen.mode.visible.y;
	xggiScreen.depth  = GT_DEPTH(xggiScreen.mode.graphtype);
	xggiScreen.bitsPerPixel = GT_SIZE(xggiScreen.mode.graphtype);

	screenInfo->formats[0].depth = 1;
	screenInfo->formats[0].bitsPerPixel = 1;
	screenInfo->formats[0].scanlinePad = BITMAP_SCANLINE_PAD;
	screenInfo->formats[1].depth = xggiScreen.depth;
	screenInfo->formats[1].bitsPerPixel = xggiScreen.bitsPerPixel;
	screenInfo->formats[1].scanlinePad = BITMAP_SCANLINE_PAD;
		
	screenInfo->numPixmapFormats = 2;

	screenInfo->imageByteOrder = IMAGE_BYTE_ORDER;
	screenInfo->bitmapScanlineUnit = BITMAP_SCANLINE_UNIT;
	screenInfo->bitmapScanlinePad = BITMAP_SCANLINE_PAD;
	screenInfo->bitmapBitOrder = BITMAP_BIT_ORDER;

	/* initialize screens */

	if (AddScreen(xggiFBInitProc, argc, argv) == -1) {
		FatalError("Couldn't add screen");
	}
}
